import gc
import os
import re
import sys
from . import threadsafe_tkinter as tk

from copy import deepcopy
from datetime import datetime
from io import StringIO
from time import time, sleep
from os.path import basename, dirname, exists, join, isfile, relpath
from tkinter.font import Font
from tkinter.constants import *
from tkinter.filedialog import askopenfilenames, askopenfilename,\
     askdirectory, asksaveasfilename
from tkinter import messagebox
from traceback import format_exc

# load the binilla constants so they are injected before any defs are loaded
from . import constants as s_c
s_c.inject()
from supyr_struct.field_types import FieldType
from supyr_struct.defs.util import *

from . import editor_constants as e_c
from .tag_window import *
from .config_def import *
from .widget_picker import *
from .widgets import BinillaWidget, ToolTipHandler
from .handler import Handler

this_curr_dir = dirname(__file__)
default_config_path = this_curr_dir + '%sbinilla.cfg' % s_c.PATHDIV

default_hotkeys = {
    '<Control-w>': 'close_selected_window',
    '<Control-o>': 'load_tags',
    '<Control-n>': 'new_tag',
    '<Control-s>': 'save_tag',
    '<Control-f>': 'show_defs',
    '<Control-p>': 'print_tag',

    '<Control-BackSpace>': 'clear_console',
    '<Control-backslash>': 'cascade',
    '<Control-Shift-bar>': 'tile_vertical',
    '<Control-Shift-underscore>': 'tile_horizontal',

    '<Alt-m>': 'minimize_all',
    '<Alt-r>': 'restore_all',
    '<Alt-w>': 'show_window_manager',
    '<Alt-c>': 'show_config_file',
    '<Alt-o>': 'load_tag_as',
    '<Alt-s>': 'save_tag_as',
    '<Alt-F4>': 'exit',

    '<Alt-Control-c>': 'apply_config',

    '<Control-Shift-s>': 'save_all',
    }


default_tag_window_hotkeys = {
    '<Control-z>': 'edit_undo',
    '<Control-y>': 'edit_redo',
    '<MouseWheel>': 'mousewheel_scroll_y',
    '<Shift-MouseWheel>': 'mousewheel_scroll_x',
    }


class IORedirecter(StringIO):
    # Text widget to output text to
    text_out = None
    log_file = None
    edit_log = None

    def __init__(self, text_out, *args, **kwargs):
        self.log_file = kwargs.pop('log_file', None)
        self.edit_log = kwargs.pop('edit_log', None)
        StringIO.__init__(self, *args, **kwargs)
        self.text_out = text_out

    def write(self, string):
        if self.edit_log and self.log_file is not None:
            try:
                self.log_file.write(string)
            except Exception:
                pass
        self.text_out.config(state=NORMAL)
        self.text_out.insert(END, string)
        self.text_out.see(END)
        self.text_out.config(state=DISABLED)


class Binilla(tk.Tk, BinillaWidget):
    # the tag of the currently in-focus TagWindow
    selected_tag = None
    # the Handler for managing loaded tags
    handler = None

    # a window that displays and allows selecting loaded definitions
    def_selector_window = None
    # a window that allows you to select a TagWindow from all open ones
    tag_window_manager = None

    # the default WidgetPicker instance to use for selecting widgets
    widget_picker = WidgetPicker()
    def_tag_window_cls = TagWindow

    # dict of open TagWindow instances. keys are the ids of each of the windows
    tag_windows = None
    # map of the id of each tag to the id of the window displaying it
    tag_id_to_window_id = None

    '''Directories'''
    curr_dir = this_curr_dir
    styles_dir = join(curr_dir, "styles")
    last_load_dir = curr_dir
    last_defs_dir = curr_dir
    last_imp_dir  = curr_dir

    recent_tag_max = 20
    recent_tagpaths = ()

    '''Miscellaneous properties'''
    _initialized = False
    app_name = "Binilla"  # the name of the app(used in window title)
    version = '0.9.40'
    log_filename = 'binilla.log'
    debug = 0
    debug_mode = False
    untitled_num = 0  # when creating a new, untitled tag, this integer is used
    #                   in its name like so: 'untitled%s' % self.untitled_num
    max_undos = 1000

    '''Config properties'''
    style_def = style_def
    config_def = config_def
    config_version = 1
    config_path = default_config_path
    config_window = None
    # the tag that holds all the config settings for this application
    config_file = None
    log_file = None

    '''Window properties'''
    # When tags are opened they are tiled, first vertically, then horizontally.
    # curr_step_y is incremented for each tag opened till it reaches max_step_y
    # At that point it resets to 0 and increments curr_step_x by 1.
    # If curr_step_x reaches max_step_x it will reset to 0. The position of an
    # opened TagWindow is set relative to the application's top left corner.
    # The x offset is shifted right by curr_step_x*tile_stride_x and
    # the y offset is shifted down  by curr_step_y*tile_stride_y.
    max_step_x = 4
    max_step_y = 8

    curr_step_x = 0
    curr_step_y = 0

    cascade_stride_x = 60
    tile_stride_x = 120
    tile_stride_y = 30

    default_tag_window_width = 480
    default_tag_window_height = 640

    window_menu_max_len = 15

    app_width = 640
    app_height = 480
    app_offset_x = 0
    app_offset_y = 0

    scroll_increment_x = 50
    scroll_increment_y = 50

    terminal_out = ''

    sync_window_movement = True  # Whether or not to sync the movement of
    #                              the TagWindow instances with the app.

    # a mapping of hotkey bindings to method names
    curr_hotkeys = None
    curr_tag_window_hotkeys = None

    # the ToolTipFrame that displays the tooltip of whatever
    # widget is currently in focus or under the mouse.
    tooltip_handler = None

    def __init__(self, *args, **kwargs):
        for s in ('curr_dir', 'config_version', 'window_menu_max_len',
                  'app_width', 'app_height', 'app_offset_x', 'app_offset_y'):
            if s in kwargs:
                object.__setattr__(self, s, kwargs.pop(s))

        self.widget_picker = kwargs.pop('widget_picker', self.widget_picker)
        self.debug = kwargs.pop('debug', self.debug)
        self.tag_windows = {}
        self.tag_id_to_window_id = {}

        if 'handler' in kwargs:
            self.handler = kwargs.pop('handler')
        else:
            self.handler = Handler(debug=self.debug)

        self.recent_tagpaths = []
        if self.curr_hotkeys is None:
            self.curr_hotkeys = {}
        if self.curr_tag_window_hotkeys is None:
            self.curr_tag_window_hotkeys = {}

        if self.config_file is not None:
            pass
        elif exists(self.config_path):
            # load the config file
            try:
                self.load_config()
            except Exception:
                print(format_exc())
                self.make_config()
        else:
            # make a config file
            self.make_config()

        if not exists(self.curr_dir):
            self.curr_dir = this_curr_dir
            try:
                self.config_file.data.directory_paths.curr_dir.path = curr_dir
            except Exception:
                pass

        self.app_name = kwargs.pop('app_name', self.app_name)
        self.app_name = str(kwargs.pop('version', self.app_name))

        if self.handler is not None:
            self.handler.log_filename = self.log_filename

        tk.Tk.__init__(self, *args, **kwargs)

        #fonts
        self.fixed_font = Font(family="Courier", size=8)
        self.container_title_font = Font(
            family="Courier", size=10, weight='bold')
        self.comment_font = Font(family="Courier", size=9)
        
        self.title('%s v%s' % (self.app_name, self.version))
        self.minsize(width=200, height=50)
        self.protocol("WM_DELETE_WINDOW", self.exit)
        self.bind('<Configure>', self.sync_tag_window_pos)

        self.bind_hotkeys()

        ######################################################################
        ######################################################################
        # MAKE METHODS FOR CREATING/DESTROYING MENUS SO THEY CAN BE CUSTOMIZED
        # This includes creating a system to manage the menus and keep track
        # of which ones exist, their order on the menu bar, their names, etc.
        # Once that is done, replace the below code with code that uses them.
        ######################################################################
        ######################################################################

        #create the main menu and add its commands
        self.main_menu = tk.Menu(self)
        self.file_menu = tk.Menu(self.main_menu, tearoff=0)
        self.edit_menu = tk.Menu(self.main_menu, tearoff=0)
        self.settings_menu = tk.Menu(self.main_menu, tearoff=0)
        self.debug_menu   = tk.Menu(self.main_menu, tearoff=0)
        self.windows_menu = tk.Menu(
            self.main_menu, tearoff=0, postcommand=self.generate_windows_menu)
        self.recent_tags_menu = tk.Menu(
            self.main_menu, tearoff=0,
            postcommand=self.generate_recent_tag_menu)

        self.config(menu=self.main_menu)

        #add cascades and commands to the main_menu
        self.main_menu.add_cascade(label="File",    menu=self.file_menu)
        #self.main_menu.add_cascade(label="Edit",   menu=self.edit_menu)
        self.main_menu.add_cascade(label="Settings", menu=self.settings_menu)
        self.main_menu.add_cascade(label="Windows", menu=self.windows_menu)
        #self.main_menu.add_command(label="Help")
        #self.main_menu.add_command(label="About")
        try:
            self.debug_mode = bool(self.config_file.data.header.flags.debug_mode)
        except Exception:
            self.debug_mode = True

        if self.debug_mode:
            self.main_menu.add_cascade(label="Debug", menu=self.debug_menu)

        #add the commands to the file_menu
        fm_ac = self.file_menu.add_command
        fm_ac(label="New",        command=self.new_tag)
        self.file_menu.add_cascade(label="Recent tags     ",
                                   menu=self.recent_tags_menu)
        fm_ac(label="Open",       command=self.load_tags)
        fm_ac(label="Open as...", command=self.load_tag_as)
        fm_ac(label="Close", command=self.close_selected_window)
        self.file_menu.add_separator()
        fm_ac(label="Save",       command=self.save_tag)
        fm_ac(label="Save as...", command=self.save_tag_as)
        fm_ac(label="Save all",   command=self.save_all)
        self.file_menu.add_separator()
        fm_ac(label="Exit",       command=self.exit)

        #add the commands to the settings_menu
        '''
        UNCOMMENT THESE WHEN THE DEFINITION LOADING ACTUALLY WORKS
        '''
        self.settings_menu.add_command(
            label="Load definitions", command=self.select_defs)
        self.settings_menu.add_command(
            label="Show definitions", command=self.show_defs)
        self.settings_menu.add_separator()
        self.settings_menu.add_command(
            label="Edit config", command=self.show_config_file)
        self.settings_menu.add_separator()
        self.settings_menu.add_command(
            label="Load style", command=self.apply_style)
        self.settings_menu.add_command(
            label="Save current style", command=self.make_style)

        self.debug_menu.add_command(label="Print tag", command=self.print_tag)
        self.debug_menu.add_command(label="Clear console",
                                    command=self.clear_console)
        self.debug_menu.add_separator()
        self.debug_menu.add_command(
            label="Force big endian", command=self.force_big_endian)
        self.debug_menu.add_command(
            label="Force little endian", command=self.force_little_endian)
        self.debug_menu.add_command(
            label="Force normal endian", command=self.force_normal_endian)
        self.debug_menu.add_separator()
        
        # make the canvas for anything in the main window
        self.root_frame = tk.Frame(self, bd=3, highlightthickness=0,
                                   relief=SUNKEN)
        self.root_frame.pack(fill=BOTH, side=LEFT, expand=True)
        self.tooltip_handler = ToolTipHandler(self)

        # make the io redirector and redirect sys.stdout to it
        self.orig_stdout = sys.stdout

        flags = self.config_file.data.header.flags
        if flags.log_output:
            curr_dir = self.curr_dir
            if not curr_dir.endswith(s_c.PATHDIV):
                curr_dir += s_c.PATHDIV

            try:
                self.log_file = open(curr_dir + self.log_filename, 'a+')

                # write a timestamp to the file
                time = datetime.now().strftime("%Y-%m-%d  %H:%M:%S")
                self.log_file.write("\n%s%s%s\n" %
                                    ("-"*30, time, "-"*(50-len(time))))
            except Exception:
                pass

        # make the console output
        self.make_io_text()

        self.geometry("%sx%s+%s+%s" %
                      (self.app_width, self.app_height,
                       self.app_offset_x, self.app_offset_y))
        self.update_window_settings()
        try:
            if self.config_file.data.header.flags.load_last_workspace:
                self.load_last_workspace()
        except Exception:
            pass
        self._initialized = True

    def add_to_recent(self, filepath):
        recent = self.recent_tagpaths
        recent.append(filepath)
        for i in range(len(recent)-1, -1, -1):
            if recent[i] == filepath:
                recent.pop(i)
        if len(recent) > self.recent_tag_max:
            del recent[0: len(recent) - self.recent_tag_max]
        recent.append(filepath)

    def add_tag(self, tag, new_filepath=''):
        '''
        new_filepath is expected to be a relative filepath if
        self.handler.tagsdir_relative == True
        '''
        filepath = tag.filepath
        handler = self.handler
        tags_dir = handler.tagsdir
        tagsdir_rel = handler.tagsdir_relative
        add_to_recent = True

        abs_filepath = tag.filepath
        abs_new_filepath = new_filepath
        if tagsdir_rel:
            abs_filepath = join(tags_dir, abs_filepath)
            abs_new_filepath = join(tags_dir, abs_new_filepath)

        try:
            existing_tag = handler.get_tag(tag.rel_filepath, tag.def_id)
        except Exception:
            existing_tag = None

        if not existing_tag:
            # tag doesnt already exist.
            # we dont need to remove it from anything.
            pass
        elif existing_tag is tag:
            # remove the tag from the handler under its current filepath
            self.delete_tag(tag, False, False)
        else:
            print('%s is already loaded' % abs_filepath)
            return False

        if not filepath:
            # the path is blank(new tag), give it a unique name
            new_filepath = 'untitled%s%s' % (self.untitled_num, tag.ext)
            abs_new_filepath = join(tags_dir, new_filepath)
            self.untitled_num += 1
            add_to_recent = False

        if abs_new_filepath:
            tag.filepath = abs_new_filepath

        if add_to_recent:
            self.add_to_recent(tag.filepath)

        tag.tags_dir = tags_dir
        if tagsdir_rel:
            tag.rel_filepath = new_filepath

        # index the tag under its new filepath
        handler.add_tag(tag, new_filepath)

        return True

    def make_io_text(self, master=None):
        if master is None:
            master = self.root_frame

        self.io_frame = tk.Frame(master, highlightthickness=0)
        self.io_text = tk.Text(self.io_frame,
                               font=self.fixed_font, state=DISABLED,
                               fg=self.io_fg_color, bg=self.io_bg_color)
        self.io_scroll_y = tk.Scrollbar(self.io_frame, orient=VERTICAL)

        self.io_scroll_y.config(command=self.io_text.yview)
        self.io_text.config(yscrollcommand=self.io_scroll_y.set)

        self.io_scroll_y.pack(fill=Y, side=RIGHT)
        self.io_text.pack(fill=BOTH, expand=True)
        self.io_frame.pack(fill=BOTH, expand=True)

        self.terminal_out = sys.stdout = IORedirecter(
            self.io_text, log_file=self.log_file,
            edit_log=self.config_file.data.header.flags.log_output)

    def bind_hotkeys(self, new_hotkeys=None):
        '''
        Binds the given hotkeys to the given methods of this class.
        Class methods must be the name of each method as a string.
        '''
        if new_hotkeys is None:
            new_hotkeys = {}
            for hotkey in self.config_file.data.hotkeys:
                combo = make_hotkey_string(hotkey)
                if combo is None or not hotkey.method.enum_name:
                    continue
                new_hotkeys[combo] = hotkey.method.enum_name
        assert isinstance(new_hotkeys, dict)

        # unbind any old hotkeys
        self.unbind_hotkeys()
        self.curr_hotkeys = new_hotkeys

        for hotkey, func_name in new_hotkeys.items():
            try:
                if hasattr(self, func_name):
                    self.bind_all(hotkey, self.__getattribute__(func_name))
            except Exception:
                print(format_exc())

    def cascade(self, e=None):
        windows = self.tag_windows

        # reset the offsets to 0 and get the strides
        self.curr_step_y = 0
        self.curr_step_x = 0
        x_stride = self.cascade_stride_x
        y_stride = self.tile_stride_y
        
        # reposition the window
        for wid in sorted(windows):
            window = windows[wid]

            # dont cascade hidden windows
            if window.state() == 'withdrawn':
                continue

            if self.curr_step_y > self.max_step_y:
                self.curr_step_y = 0
                self.curr_step_x += 1
            if self.curr_step_x > self.max_step_x:
                self.curr_step_x = 0

            self.place_window_relative(
                window, (self.curr_step_x * x_stride +
                         self.curr_step_y * (x_stride//2) + 5),
                self.curr_step_y*y_stride + 50)
            self.curr_step_y += 1
            window.update_idletasks()

            self.selected_tag = None
            self.select_tag_window(window)

    def close_selected_window(self, e=None):
        if self.selected_tag is None:
            return

        # close the window and make sure the tag is deleted
        self.delete_tag(self.selected_tag)

        # select the next TagWindow
        try:
            for w in self.tag_windows.values():
                if hasattr(w, 'tag') and w.state() != 'withdrawn':
                    self.select_tag_window(w)
                    self.update()
                    return
        except Exception:
            pass

    def clear_console(self, e=None):
        try:
            self.io_text.config(state=NORMAL)
            self.io_text.delete('1.0', END)
            self.io_text.config(state=DISABLED)
        except Exception:
            print(format_exc())

    def delete_tag(self, tag, destroy_window=True, forget_window=True):
        try:
            if tag is self.config_file:
                pass
            elif hasattr(tag, "rel_filepath"):
                # remove the tag from the handlers tag library.
                # We need to delete it by the relative filepath
                # rather than having it detect it using the tag
                # because the handlers tagsdir may have changed
                # from what it was when the tag was created, so
                # it wont be able to determine the rel_filepath
                tag.handler.delete_tag(filepath=tag.rel_filepath)
            else:
                tag.handler.delete_tag(filepath=tag.filepath)

            if destroy_window or forget_window:
                tid = id(tag)
                def_id = tag.def_id
                tid_to_wid = self.tag_id_to_window_id

                if tid in tid_to_wid:
                    wid = tid_to_wid[tid]
                    t_window = self.tag_windows[wid]
                    del tid_to_wid[tid]
                    del self.tag_windows[wid]

                    if destroy_window:
                        t_window.destroy()

            if self.selected_tag is tag:
                self.selected_tag = None

            gc.collect()
        except Exception:
            print(format_exc())

    def exit(self, e=None):
        '''Exits the program.'''
        try:
            sys.stdout = self.orig_stdout
            if self.log_file:
                self.log_file.close()
        except Exception:
            print(format_exc())

        try:
            self.record_open_tags()
            self.update_config()
        except Exception:
            print(format_exc())

        windows = ()
        try: windows += tuple(self.tag_windows.values())
        except Exception: print(format_exc())
        try: windows += tuple(self.children.values())
        except Exception: print(format_exc())

        try:
            # need to save before destroying the
            # windows or bindings wont be saved
            self.save_config()
        except Exception:
            print(format_exc())

        for w in windows:
            try:
                not_destroyed = bool(w.destroy())
                if not_destroyed:
                    return
            except Exception:
                pass

        self.tk.unhook()

        try: self.destroy()  # wont close if a listener is open without this
        except Exception: pass

    def force_big_endian(self, e=None):
        FieldType.force_big()

    def force_little_endian(self, e=None):
        FieldType.force_little()

    def force_normal_endian(self, e=None):
        FieldType.force_normal()

    def record_open_tags(self):
        try:
            handler = self.handler
            config_file = self.config_file
            open_tags = config_file.data.open_tags
            del open_tags[:]

            for w in self.tag_windows.values():
                tag = w.tag

                # dont store tags that arent from the current handler
                if tag in (config_file, None) or tag.handler is not handler:
                    continue

                open_tags.append()
                open_tag = open_tags[-1]
                open_header = open_tag.header

                if w.state() == 'withdrawn':
                    open_header.flags.minimized = True

                pos_x, pos_y = w.winfo_x(), w.winfo_y()
                width, height = w.geometry().split('+')[0].split('x')[:2]

                open_header.offset_x, open_header.offset_y = pos_x, pos_y
                open_header.width, open_header.height = int(width), int(height)

                open_tag.def_id = tag.def_id
                open_tag.path = tag.filepath
        except Exception:
            print(format_exc())

    def load_last_workspace(self):
        try:
            handler = self.handler
            config_file = self.config_file
            open_tags = config_file.data.open_tags

            for open_tag in open_tags:
                def_id = open_tag.def_id
                path = open_tag.path

                open_header = open_tag.header

                windows = self.load_tags(filepaths=path, def_id=def_id)
                if not windows:
                    continue

                w = windows[0]
                if open_header.flags.minimized:
                    w.withdraw()
                    self.selected_tag = None

                pos_x, pos_y = open_header.offset_x, open_header.offset_y
                width, height = open_header.width, open_header.height

                w.geometry("%sx%s+%s+%s" % (width, height, pos_x, pos_y))
        except Exception:
            print(format_exc())

    def generate_windows_menu(self):
        menu = self.windows_menu
        menu.delete(0, "end")  # clear the menu
        sync_word = 'off' if self.sync_window_movement else 'on'

        #add the commands to the windows_menu
        menu.add_command(label="Minimize all", command=self.minimize_all)
        menu.add_command(label="Restore all", command=self.restore_all)
        menu.add_command(label="Turn movement sync " + sync_word,
                         command=self.toggle_sync)
        menu.add_separator()
        menu.add_command(label="Cascade", command=self.cascade)
        menu.add_command(label="Tile vertical", command=self.tile_vertical)
        menu.add_command(label="Tile horizontal", command=self.tile_horizontal)

        i = 0
        max_len = self.window_menu_max_len

        if not self.tag_windows:
            return

        menu.add_separator()

        # store the windows by label
        windows_by_label = {}
        for w in self.tag_windows.values():
            windows_by_label[w.title()] = w

        for label in sorted(windows_by_label):
            w = windows_by_label[label]
            if i >= max_len:
                menu.add_separator()
                menu.add_command(label="Window manager",
                                 command=self.show_window_manager)
                break
            try:
                menu.add_command(
                    label=label, command=lambda w=w: self.select_tag_window(w))
                i += 1
            except Exception:
                print(format_exc())

    def generate_recent_tag_menu(self):
        menu = self.recent_tags_menu
        menu.delete(0, "end")  # clear the menu

        i = 1
        menu.add_command(label="Clear recently opened tags",
                         command=lambda self=self:
                         self.recent_tagpaths.__delitem__(
                             slice(None, None, None))
                         )
        menu.add_separator()
        for tagpath in reversed(self.recent_tagpaths):
            try:
                menu.add_command(
                    label="%s  %s" % (i, tagpath),
                    command=lambda s=tagpath: self.load_tags(s))
                i += 1
            except Exception:
                print(format_exc())

    def apply_config(self, e=None):
        config_data = self.config_file.data
        header = config_data.header
        app_window = config_data.app_window

        open_tags = config_data.open_tags
        recent_tags = config_data.recent_tags
        dir_paths = config_data.directory_paths

        self.sync_window_movement = header.flags.sync_window_movement

        __osa__ = object.__setattr__
        __tsa__ = type.__setattr__

        if self._initialized:
            self.bind_hotkeys()
        else:
            # only load the recent tagpaths when loading binilla
            self.recent_tagpaths = paths = []

            for tagpath in recent_tags:
                paths.append(tagpath.path)

        for s in app_window.NAME_MAP.keys():
            try: __osa__(self, s, app_window[s])
            except IndexError: pass

        for s in ('recent_tag_max', 'max_undos'):
            try: __osa__(self, s, header[s])
            except IndexError: pass

        for s in ('last_load_dir', 'last_defs_dir', 'last_imp_dir',
                  'curr_dir', 'styles_dir')[:len(dir_paths)]:
            try: __osa__(self, s, dir_paths[s].path)
            except IndexError: pass

        for w in self.tag_windows.values():
            try: w.edit_resize(self.max_undos)
            except Exception: print(format_exc())

        self.handler.tagsdir = dir_paths.tags_dir.path

        self.log_filename = basename(dir_paths.debug_log_path.path)

        try:
            self.debug_mode = bool(header.flags.debug_mode)
        except Exception:
            self.debug_mode = True

        self.apply_style(style_file=self.config_file)

    def load_config(self, filepath=None):
        if filepath is None:
            filepath = self.config_path
        assert exists(filepath)

        # load the config file
        self.config_file = self.config_def.build(filepath=filepath)
        if self.config_file.data.header.version != self.config_version:
            raise ValueError(
                "Config version is not what this application is expecting.")
        app_window = self.config_file.data.app_window

        self.app_width = app_window.app_width
        self.app_height = app_window.app_height
        self.app_offset_x = app_window.app_offset_x
        self.app_offset_y = app_window.app_offset_y

        self.apply_config()

        hotkeys = self.config_file.data.hotkeys
        tag_window_hotkeys = self.config_file.data.tag_window_hotkeys

        for hotkey in hotkeys:
            combo = make_hotkey_string(hotkey)
            if combo is None or not hotkey.method.enum_name:
                continue
            self.curr_hotkeys[combo] = hotkey.method.enum_name

        for hotkey in tag_window_hotkeys:
            combo = make_hotkey_string(hotkey)
            if combo is None or not hotkey.method.enum_name:
                continue
            self.curr_tag_window_hotkeys[combo] = hotkey.method.enum_name

    def apply_style(self, filepath=None, style_file=None):
        if isinstance(filepath, tk.Event):
            filepath = None
        if style_file is None:
            if filepath is None:
                filepath = askopenfilename(
                    initialdir=self.styles_dir, parent=self,
                    title="Select style to load",
                    filetypes=(("binilla_style", "*.sty"), ('All', '*')))

            if not filepath:
                return

            assert exists(filepath)
            self.styles_dir = dirname(filepath)
            style_file = self.style_def.build(filepath=filepath)

        assert hasattr(style_file, 'data')

        style_data = style_file.data

        header = style_data.header
        widgets = style_data.widgets

        widget_depths = widgets.depths
        colors = style_data.colors

        __osa__ = object.__setattr__
        __tsa__ = type.__setattr__

        for s in ('title_width', 'scroll_menu_width', 'enum_menu_width',
                  'min_entry_width', 'textbox_width', 'textbox_height',
                  'bool_frame_min_width', 'bool_frame_min_height',
                  'bool_frame_max_width', 'bool_frame_max_height',
                  'def_int_entry_width',    'max_int_entry_width',
                  'def_float_entry_width',  'max_float_entry_width', 
                  'def_string_entry_width', 'max_string_entry_width',
                  'scroll_menu_max_width', 'scroll_menu_max_height', ):
            try: __tsa__(BinillaWidget, s, widgets[s])
            except IndexError: pass

        for s in ('vertical_padx', 'vertical_pady',
                  'horizontal_padx', 'horizontal_pady'):
            try: __tsa__(BinillaWidget, s, tuple(widgets[s]))
            except IndexError: pass

        for s in widget_depth_names[:len(widget_depths)]:
            try: __tsa__(BinillaWidget, s + '_depth', widget_depths[s])
            except IndexError: pass

        for s in color_names[:len(colors)]:
            # it has to be a tuple for some reason
            try:
                __tsa__(BinillaWidget, s + '_color',
                        '#%02x%02x%02x' % tuple(colors[s]))
            except IndexError:
                pass

        if self._initialized:
            self.update_config()
            self.update_window_settings()
            self.geometry("%sx%s+%s+%s" %
                          (self.app_width, self.app_height,
                           self.app_offset_x, self.app_offset_y))

    def make_config(self, filepath=None):
        if filepath is None:
            filepath = self.config_path

        # create the config file from scratch
        self.config_file = self.config_def.build()
        self.config_file.filepath = filepath

        data = self.config_file.data

        # make sure these have as many entries as they're supposed to
        for block in (data.directory_paths, data.widgets.depths, data.colors):
            block.extend(len(block.NAME_MAP))

        self.curr_hotkeys = dict(default_hotkeys)
        self.curr_tag_window_hotkeys = dict(default_tag_window_hotkeys)

        self.update_config()

        c_hotkeys = data.hotkeys
        c_tag_window_hotkeys = data.tag_window_hotkeys

        for k_set, b in ((default_hotkeys, c_hotkeys),
                         (default_tag_window_hotkeys, c_tag_window_hotkeys)):
            default_keys = k_set
            hotkeys = b
            for combo, method in k_set.items():
                hotkeys.append()
                keys = hotkeys[-1].combo

                modifier, key = read_hotkey_string(combo)
                keys.modifier.set_to(modifier)
                keys.key.set_to(key)

                hotkeys[-1].method.set_to(method)

    def make_style(self):
        # create the style file from scratch
        filepath = asksaveasfilename(
            initialdir=self.styles_dir, defaultextension='.sty',
            title="Save style as...", parent=self,
            filetypes=(("binilla style", "*.sty"), ('All', '*')))

        if not filepath:
            return

        self.styles_dir = dirname(filepath)
        style_file = self.style_def.build()
        style_file.filepath = filepath

        style_file.data.widgets.depths.extend(len(widget_depth_names))
        style_file.data.colors.extend(len(color_names))

        self.update_style(style_file)
        style_file.serialize(temp=0, backup=0, calc_pointers=0)

    def toggle_sync(self):
        self.config_file.data.header.flags.sync_window_movement = (
            self.sync_window_movement) = not self.sync_window_movement

    def get_tag(self, filepath, def_id=None):
        '''
        Returns the tag from the handler under the given def_id and filepath.
        filepath is expected to be relative to self.tags_dir
        '''
        if def_id is None:
            def_id = self.handler.get_def_id(filepath)
        try:
            return self.handler.get_tag(sanitize_path(filepath), def_id)
        except Exception:
            pass

    def get_tag_window_by_tag(self, tag):
        return self.tag_windows[self.tag_id_to_window_id[id(tag)]]

    def get_is_tag_loaded(self, filepath, def_id=None):
        if def_id is None:
            def_id = self.handler.get_def_id(filepath)
        return bool(self.get_tag(filepath, def_id))

    def load_tags(self, filepaths=None, def_id=None):
        '''Prompts the user for a tag(s) to load and loads it.'''
        if isinstance(filepaths, tk.Event):
            filepaths = None
        if filepaths is None:
            filetypes = [('All', '*')]
            defs = self.handler.defs
            for id in sorted(defs.keys()):
                filetypes.append((id, defs[id].ext))
            filepaths = askopenfilenames(initialdir=self.last_load_dir,
                                         filetypes=filetypes, parent=self,
                                         title="Select the tag to load")
            if not filepaths:
                return

        if isinstance(filepaths, str):
            # account for a stupid bug with certain versions of windows
            if filepaths.startswith('{'):
                filepaths = re.split("\}\W\{", filepaths[1:-1])
            else:
                filepaths = (filepaths, )

        self.last_load_dir = dirname(filepaths[-1])
        w = None

        windows = []
        handler = self.handler
        handler_flags = self.config_file.data.header.handler_flags
        tags_dir = join(handler.tagsdir, "")

        for path in filepaths:
            path = abs_path = sanitize_path(path)
            is_new_tag = not bool(abs_path)

            if self.get_is_tag_loaded(path):
                # the tag is somehow still loaded.
                # need to see if there is still a window
                new_tag = self.get_tag(path, handler.get_def_id(path))
                try:
                    w = self.get_tag_window_by_tag(new_tag)
                except Exception:
                    w = None
                if w:
                    print('%s is already loaded' % path)
                    continue

                # there isn't a window, so continue like normal
            else:
                # try to load the new tags
                try:
                    if handler.tagsdir_relative and not is_new_tag:
                        abs_path = join(tags_dir, abs_path)
                    new_tag = handler.build_tag(
                        filepath=abs_path, def_id=def_id,
                        allow_corrupt=handler_flags.allow_corrupt)
                except Exception:
                    if handler.debug:
                        print(format_exc())
                    print("Could not load tag '%s'" % path)
                    continue

            self.add_tag(new_tag, path)

            try:
                #build the window
                w = self.make_tag_window(new_tag, focus=False,
                                         is_new_tag=is_new_tag)
                windows.append(w)
            except Exception:
                print(format_exc())
                raise IOError("Could not display tag '%s'." % path)

        self.select_tag_window(w)
        return windows

    def load_tag_as(self, e=None):
        '''Prompts the user for a tag to load and loads it.'''
        if self.def_selector_window:
            return
        
        filetypes = [('All', '*')]
        defs = self.handler.defs
        for def_id in sorted(defs.keys()):
            filetypes.append((def_id, defs[def_id].ext))
        fp = askopenfilename(initialdir=self.last_load_dir,
                             filetypes=filetypes, parent=self,
                             title="Select the tag to load")

        if not fp:
            return

        self.last_load_dir = dirname(fp)
        dsw = DefSelectorWindow(
            self, title="Select a definition to use", action=lambda def_id:
            self.load_tags(filepaths=fp, def_id=def_id))
        self.def_selector_window = dsw
        self.place_window_relative(self.def_selector_window, 30, 50)

    def place_window_relative(self, window, x=0, y=0):
        # calculate x and y coordinates for this window
        x_base, y_base = self.winfo_x(), self.winfo_y()
        w, h = window.geometry().split('+')[0].split('x')[:2]
        if w == h and w == '1':
            w = window.winfo_reqwidth()
            h = window.winfo_reqheight()
        window.geometry('%sx%s+%s+%s' % (w, h, x + x_base, y + y_base))

    def make_tag_window(self, tag, *, focus=True, window_cls=None,
                        is_new_tag=False):
        '''
        Creates and returns a TagWindow instance for the supplied
        tag and sets the current focus to the new TagWindow.
        '''
        if len(self.tag_windows) == 0:
            self.curr_step_y = self.curr_step_x = 0
        if window_cls is None:
            window_cls = self.def_tag_window_cls
        window = window_cls(
            self, tag, app_root=self, handler=self.handler,
            is_new_tag=is_new_tag, widget_picker=self.widget_picker)

        # reposition the window
        if self.curr_step_y > self.max_step_y:
            self.curr_step_y = 0
            self.curr_step_x += 1
        if self.curr_step_x > self.max_step_x:
            self.curr_step_x = 0

        self.place_window_relative(
            window, self.curr_step_x*self.tile_stride_x + 5,
            self.curr_step_y*self.tile_stride_y + 50)
        self.curr_step_y += 1

        try:
            use_default_size = self.config_file.data.header.tag_window_flags.\
                               use_default_window_dimensions
        except Exception:
            use_default_size = False

        if use_default_size:
            window.geometry("%sx%s" % (
                self.default_tag_window_width, self.default_tag_window_height))

        self.tag_windows[id(window)] = window
        self.tag_id_to_window_id[id(tag)] = id(window)

        # make sure the window is drawn now that it exists
        window.update_idletasks()

        if focus:
            # set the focus to the new TagWindow
            self.select_tag_window(window)

        return window

    def minimize_all(self, e=None):
        '''Minimizes all open TagWindows.'''
        windows = self.tag_windows
        for wid in sorted(windows):
            w = windows[wid]
            try:
                w.withdraw()
            except Exception:
                print(format_exc())

    def new_tag(self, e=None):
        if self.def_selector_window:
            return
        
        dsw = DefSelectorWindow(
            self, title="Select a definition to use", action=lambda def_id:
            self.load_tags(filepaths='', def_id=def_id))
        self.def_selector_window = dsw
        self.place_window_relative(self.def_selector_window, 30, 50)

    def print_tag(self, e=None):
        '''Prints the currently selected tag to the console.'''
        try:
            if self.selected_tag is None:
                return
            try:
                show = set()
                header = self.config_file.data.header
                precision = header.print_precision
                indent = header.print_indent

                for name in header.block_print.NAME_MAP:
                    if header.block_print.get(name):
                        show.add(name.split('show_')[-1])

                if not header.flags.log_tag_print:
                    self.terminal_out.edit_log = False
            except Exception:
                show = s_c.MOST_SHOW

            tag_str = self.selected_tag.pprint(
                show=show, precision=precision, indent=indent)

            # print the string line by line
            for line in tag_str.split('\n'):
                try:
                    print(line)
                except:
                    print(' '*(len(line)-len(line.lstrip(' ')))+s_c.UNPRINTABLE)
                self.io_text.update()

            try: self.terminal_out.edit_log = bool(header.flags.log_output)
            except Exception: pass
        except Exception:
            print(format_exc())

    def restore_all(self, e=None):
        '''Restores all open TagWindows to being visible.'''
        windows = self.tag_windows
        for wid in sorted(windows):
            w = windows[wid]
            try:
                if w.state() == 'withdrawn':
                    w.deiconify()
            except Exception:
                print(format_exc())

    def save_config(self, e=None):
        self.config_file.serialize(temp=0, backup=0, calc_pointers=0)
        self.apply_config()

    def save_tag(self, tag=None):
        if isinstance(tag, tk.Event):
            tag = None
        if tag is None:
            if self.selected_tag is None:
                return
            tag = self.selected_tag

        if tag is self.config_file:
            return self.save_config()

        if hasattr(tag, "serialize"):
            # make sure the tag has a valid filepath whose directories
            # can be made if they dont already exist(dirname must not be '')
            if not(hasattr(tag, "filepath") and tag.filepath and
                   dirname(tag.filepath)):
                self.save_tag_as(tag)
                return

            # make sure to flush any changes made using widgets to the tag
            try:
                self.get_tag_window_by_tag(tag).save()
            except Exception:
                print(format_exc())
                raise IOError("Could not save tag.")

            path = tag.filepath
            self.add_to_recent(path)

        return tag

    def save_tag_as(self, tag=None, filepath=None):
        if isinstance(tag, tk.Event):
            tag = None
        if tag is None:
            if self.selected_tag is None:
                return
            tag = self.selected_tag

        if not hasattr(tag, "serialize"):
            return

        if filepath is None:
            ext = tag.ext
            filepath = asksaveasfilename(
                initialdir=dirname(tag.filepath), defaultextension=ext,
                title="Save tag as...", filetypes=[
                    (ext[1:], "*" + ext), ('All', '*')] )

        if not filepath:
            return

        # make sure to flush any changes made using widgets to the tag
        w = self.get_tag_window_by_tag(tag)

        try:
            self.last_load_dir = dirname(filepath)
            if tag.handler.tagsdir_relative:
                filepath = relpath(filepath, tags_dir)

            self.add_tag(tag, filepath)

            w.save(temp=False)
        except Exception:
            print(format_exc())
            raise IOError("Could not save tag.")

        try:
            self.get_tag_window_by_tag(tag).update_title()
        except Exception:
            # this isnt really a big deal
            #print(format_exc())
            pass

        return tag

    def save_all(self, e=None):
        '''
        Saves all currently loaded tags to their files.
        '''
        tags = self.handler.tags
        for def_id in tags:
            tag_coll = tags[def_id]
            for tag_path in tag_coll:
                try:
                    self.save_tag(tag_coll[tag_path])
                except Exception:
                    print(format_exc())
                    print("Exception occurred while trying to save '%s'" %
                          tag_path)

    def select_tag_window(self, window=None):
        try:
            if window is None:
                self.selected_tag = None
                self.focus_set()
                return

            if window.tag is not None:
                tag = window.tag
                # if the window IS selected, minimize it
                if self.selected_tag == tag:
                    self.selected_tag = None
                    return

                self.selected_tag = window.tag
                if window.state() == 'withdrawn':
                    window.deiconify()

                # focus_set wasnt working, so i had to play hard ball
                window.focus_force()
        except Exception:
            print(format_exc())

    def select_tag_window_by_tag(self, tag=None):
        if tag is None:
            return
        try:
            self.select_tag_window(self.get_tag_window_by_tag(tag))
        except Exception:
            print(format_exc())

    def select_defs(self):
        '''Prompts the user to specify where to load the tag defs from.
        Reloads the tag definitions from the folder specified.'''
        defs_dir = askdirectory(initialdir=self.last_defs_dir, parent=self,
                                title="Select the tag definitions folder")
        if defs_dir != "":
            print('Loading selected definitions...')
            self.update_idletasks()
            if not defs_dir.endswith(s_c.PATHDIV):
                defs_dir += s_c.PATHDIV
            defs_dir = sanitize_path(dirname(defs_dir))

            # try and find the module_root
            mod_root = defs_dir
            parent_dir = mod_root
            while isfile(parent_dir + s_c.PATHDIV + "__init__.py"):
                mod_root = parent_dir
                parent_dir = dirname(parent_dir)

            mod_root = dirname(parent_dir)

            # if the module_root isnt in sys.path, we need to add it so
            # the importer can resolve the import path for the definitions
            mod_root = sanitize_path(mod_root)
            if mod_root not in sys.path:
                sys.path.insert(0, mod_root)

            import_path = defs_dir.split(mod_root + s_c.PATHDIV)[-1]\
                          .replace(s_c.PATHDIV, ".")
            print("    Module path:  %s\n    Import path:  %s" % (
                mod_root, import_path))
            try:
                self.handler.reload_defs(defs_path=import_path)
                self.last_defs_dir = defs_dir
                print('Selected definitions loaded\n')
            except Exception:
                raise IOError("Could not load tag definitions\n.")

    def show_config_file(self, e=None):
        if self.config_window is not None:
            return

        # update the config file's directory paths
        dir_paths = self.config_file.data.directory_paths
        __oga__ = object.__getattribute__
        for s in ('last_load_dir', 'last_defs_dir', 'last_imp_dir',
                  'curr_dir', 'styles_dir', ):
            try: dir_paths[s].path = __oga__(self, s)
            except IndexError: pass

        self.config_window = self.make_tag_window(self.config_file,
                                                  window_cls=ConfigWindow)

    def show_defs(self, e=None):
        if self.def_selector_window:
            return
        
        self.def_selector_window = DefSelectorWindow(self, action=lambda x: x)
        self.place_window_relative(self.def_selector_window, 30, 50)

    def show_window_manager(self, e=None):
        if self.tag_window_manager is not None:
            return

        self.tag_window_manager = TagWindowManager(self)
        self.place_window_relative(self.tag_window_manager, 30, 50)

    def sync_tag_window_pos(self, e):
        '''Syncs TagWindows to move with the app.'''
        dx = int(self.winfo_x()) - self.app_offset_x
        dy = int(self.winfo_y()) - self.app_offset_y
        self.app_offset_x += dx
        self.app_offset_y += dy

        # keep a tabs on these so the config file can be updated
        self.app_width = self.winfo_width()
        self.app_height = self.winfo_height()

        if not self.sync_window_movement:
            return

        for w in self.tag_windows.values():
            w.geometry('%sx%s+%s+%s' %
                       (w.winfo_width(), w.winfo_height(),
                        dx + int(w.winfo_x()), dy + int(w.winfo_y())))

    def tile_vertical(self, e=None):
        windows = self.tag_windows

        # reset the offsets to 0 and get the strides
        self.curr_step_y = 0
        self.curr_step_x = 0
        x_stride = self.tile_stride_x
        y_stride = self.tile_stride_y

        # reposition the window
        for wid in sorted(windows):
            window = windows[wid]

            # dont tile hidden windows
            if window.state() == 'withdrawn':
                continue

            if self.curr_step_y > self.max_step_y:
                self.curr_step_y = 0
                self.curr_step_x += 1
            if self.curr_step_x > self.max_step_x:
                self.curr_step_x = 0

            self.place_window_relative(window,
                                       self.curr_step_x*x_stride + 5,
                                       self.curr_step_y*y_stride + 50)
            self.curr_step_y += 1
            window.update_idletasks()

            self.selected_tag = None
            self.select_tag_window(window)

    def tile_horizontal(self, e=None):
        windows = self.tag_windows

        # reset the offsets to 0 and get the strides
        self.curr_step_y = 0
        self.curr_step_x = 0
        x_stride = self.tile_stride_x
        y_stride = self.tile_stride_y

        # reposition the window
        for wid in sorted(windows):
            window = windows[wid]

            # dont tile hidden windows
            if window.state() == 'withdrawn':
                continue

            if self.curr_step_x > self.max_step_x:
                self.curr_step_x = 0
                self.curr_step_y += 1
            if self.curr_step_y > self.max_step_y:
                self.curr_step_y = 0

            self.place_window_relative(window,
                                       self.curr_step_x*x_stride + 5,
                                       self.curr_step_y*y_stride + 50)
            self.curr_step_x += 1
            window.update_idletasks()

            self.selected_tag = None
            self.select_tag_window(window)

    def unbind_hotkeys(self, hotkeys=None):
        if hotkeys is None:
            hotkeys = self.curr_hotkeys
        if isinstance(hotkeys, dict):
            hotkeys = hotkeys.keys()

        for key in tuple(hotkeys):
            try:
                self.unbind_all(key)
            except Exception:
                pass

    def update_config(self, config_file=None):
        if config_file is None:
            config_file = self.config_file
        config_data = config_file.data

        header = config_data.header

        open_tags = config_data.open_tags
        recent_tags = config_data.recent_tags
        dir_paths = config_data.directory_paths
        app_window = config_data.app_window

        header.version = self.config_version
        header.flags.sync_window_movement = self.sync_window_movement

        __oga__ = object.__getattribute__

        del recent_tags[:]

        if self._initialized:
            self.app_width = self.winfo_width()
            self.app_height = self.winfo_height()
            self.app_offset_x = self.winfo_x()
            self.app_offset_y = self.winfo_y()

            for s in app_window.NAME_MAP.keys():
                try: app_window[s] = __oga__(self, s)
                except IndexError: pass

        # make sure there are enough tagsdir entries in the directory_paths
        if len(dir_paths.NAME_MAP) > len(dir_paths):
            dir_paths.extend(len(dir_paths.NAME_MAP) - len(dir_paths))

        for path in self.recent_tagpaths:
            recent_tags.append()
            recent_tags[-1].path = path

        for s in ('recent_tag_max', 'max_undos'):
            try: header[s] = __oga__(self, s)
            except IndexError: pass

        for s in ('last_load_dir', 'last_defs_dir', 'last_imp_dir',
                  'curr_dir', 'styles_dir', ):
            try: dir_paths[s].path = __oga__(self, s)
            except IndexError: pass

        dir_paths.tags_dir.path = self.handler.tagsdir
        dir_paths.debug_log_path.path = self.log_filename

        self.update_style(config_file)

    def update_style(self, style_file):
        style_data = style_file.data
        config_data = self.config_file.data

        header = style_data.header
        widgets = style_data.widgets

        widget_depths = widgets.depths
        colors = style_data.colors

        header.parse(attr_index='date_modified')

        __oga__ = object.__getattribute__
        __tga__ = type.__getattribute__

        for s in ('title_width', 'scroll_menu_width', 'enum_menu_width',
                  'min_entry_width', 'textbox_width', 'textbox_height',
                  'bool_frame_min_width', 'bool_frame_min_height',
                  'bool_frame_max_width', 'bool_frame_max_height',
                  'def_int_entry_width',    'max_int_entry_width',
                  'def_float_entry_width',  'max_float_entry_width', 
                  'def_string_entry_width', 'max_string_entry_width',
                  'scroll_menu_max_width', 'scroll_menu_max_height', ):
            try: widgets[s] = __tga__(BinillaWidget, s)
            except IndexError: pass

        for s in ('vertical_padx', 'vertical_pady',
                  'horizontal_padx', 'horizontal_pady'):
            try: widgets[s][:] = tuple(__tga__(BinillaWidget, s))
            except IndexError: pass

        for s in widget_depth_names:
            try: widget_depths[s] = __tga__(BinillaWidget, s + '_depth')
            except IndexError: pass

        for s in color_names:
            try:
                color = __tga__(BinillaWidget, s + '_color')[1:]
                colors[s][0] = int(color[0:2], 16)
                colors[s][1] = int(color[2:4], 16)
                colors[s][2] = int(color[4:6], 16)
            except IndexError:
                pass

    def update_window_settings(self):
        for m in (self.main_menu, self.file_menu, self.settings_menu,
                  self.debug_menu, self.windows_menu, self.recent_tags_menu):
            m.config(bg=self.default_bg_color, fg=self.text_normal_color)
        self.config(bg=self.default_bg_color)
        self.io_text.config(fg=self.io_fg_color, bg=self.io_bg_color)


class DefSelectorWindow(tk.Toplevel, BinillaWidget):

    def __init__(self, app_root, action, *args, **kwargs):
        self.app_root = app_root
        try:
            title = app_root.handler.defs_filepath
        except AttributeError:
            title = "Tag definitions"

        title = kwargs.pop('title', title)
        tk.Toplevel.__init__(self, app_root, *args, **kwargs)
        self.title(title)

        self.action = action
        self.def_id = None
        self.sorted_def_ids = []
        self.minsize(width=400, height=300)

        self.list_canvas = tk.Canvas(
            self, highlightthickness=0, bg=self.default_bg_color)
        self.button_canvas = tk.Canvas(
            self, height=50, highlightthickness=0, bg=self.default_bg_color)

        #create and set the y scrollbar for the canvas root
        self.def_listbox = tk.Listbox(
            self.list_canvas, selectmode=SINGLE, highlightthickness=0,
            bg=self.enum_normal_color, fg=self.text_normal_color,
            selectbackground=self.entry_highlighted_color,
            selectforeground=self.text_highlighted_color,
            font=app_root.fixed_font)

        btn_kwargs = dict(
            bg=self.button_color, activebackground=self.button_color,
            fg=self.text_normal_color, bd=self.button_depth,
            disabledforeground=self.text_disabled_color,
            )
        self.ok_btn = tk.Button(
            self.button_canvas, text='OK', command=self.complete_action,
            width=16, **btn_kwargs)
        self.cancel_btn = tk.Button(
            self.button_canvas, text='Cancel', command=self.destroy,
            width=16, **btn_kwargs)
        self.hsb = tk.Scrollbar(self.button_canvas, orient='horizontal')
        self.vsb = tk.Scrollbar(self.list_canvas,   orient='vertical')

        self.def_listbox.config(xscrollcommand=self.hsb.set,
                                yscrollcommand=self.vsb.set)

        self.hsb.config(command=self.def_listbox.xview)
        self.vsb.config(command=self.def_listbox.yview)

        self.list_canvas.pack(fill='both', expand=True)
        self.button_canvas.pack(fill='x')

        self.vsb.pack(side=RIGHT, fill='y')
        self.def_listbox.pack(fill='both', expand=True)

        self.hsb.pack(side=TOP, fill='x')
        self.ok_btn.pack(side=LEFT,      padx=9)
        self.cancel_btn.pack(side=RIGHT, padx=9)

        # make selecting things more natural
        self.def_listbox.bind('<<ListboxSelect>>', self.set_selected_def)
        self.def_listbox.bind('<Return>', self.complete_action)
        self.def_listbox.bind('<Double-Button-1>', self.complete_action)
        self.ok_btn.bind('<Return>', self.complete_action)
        self.cancel_btn.bind('<Return>', self.destroy)
        self.bind('<Escape>', self.destroy)

        self.transient(self.app_root)
        self.grab_set()

        self.cancel_btn.focus_set()
        self.populate_listbox()

    def destroy(self, e=None):
        try:
            self.app_root.def_selector_window = None
        except AttributeError:
            pass
        tk.Toplevel.destroy(self)

    def complete_action(self, e=None):
        if self.def_id is not None:
            self.action(self.def_id)
        self.destroy()

    def populate_listbox(self):
        defs_root = self.app_root.handler.defs_path
        defs = self.app_root.handler.defs

        id_pad = ext_pad = 0
        defs_by_ext = {}

        #loop over all the defs and find the max amount of
        #padding needed between the ID and the Ext strings
        for def_id in defs:
            d = defs[def_id]
            if len(def_id) > id_pad:
                id_pad = len(def_id)

        for def_id in defs.keys():
            definiton = defs[def_id]
            ext = definiton.ext[1:]
            local_defs = defs_by_ext.get(ext, {})
            local_defs[def_id] = definiton

            defs_by_ext[ext] = local_defs

        sorted_ids = []
        for ext in sorted(defs_by_ext.keys()):
            sorted_ids.extend(tuple(defs_by_ext[ext].keys()))

        self.sorted_def_ids = tuple(sorted_ids)

        #loop over all the definitions
        for def_id in sorted_ids:
            d = defs[def_id]

            self.def_listbox.insert(END, 'ID=%s  %sExt=%s'%
                                    (def_id, ' '*(id_pad-len(def_id)),
                                     d.ext[1:] ))

    def set_selected_def(self, event=None):
        indexes = [int(i) for i in self.def_listbox.curselection()]
        
        if len(indexes) == 1:
            self.def_id = self.sorted_def_ids[indexes[0]]


class TagWindowManager(tk.Toplevel, BinillaWidget):

    app_root = None

    list_index_to_window = None

    def __init__(self, app_root, *args, **kwargs):
        self.app_root = app_root
        tk.Toplevel.__init__(self, app_root, *args, **kwargs)

        self.list_index_to_window = []

        self.title("Tag window manager")
        self.minsize(width=400, height=250)

        # make the frames
        self.windows_frame = tk.Frame(self, bg=self.default_bg_color)
        self.button_frame = tk.Frame(self, bg=self.default_bg_color)
        self.ok_frame = tk.Frame(
            self.button_frame, bg=self.default_bg_color)
        self.cancel_frame = tk.Frame(
            self.button_frame, bg=self.default_bg_color)

        btn_kwargs = dict(
            bg=self.button_color, activebackground=self.button_color,
            fg=self.text_normal_color, bd=self.button_depth,
            disabledforeground=self.text_disabled_color,
            )
        # make the buttons
        self.ok_button = tk.Button(
            self.ok_frame, text='OK', width=15,
            command=self.select, **btn_kwargs)
        self.cancel_button = tk.Button(
            self.cancel_frame, text='Cancel', width=15,
            command=self.destroy, **btn_kwargs)

        # make the scrollbars and listbox
        self.scrollbar_y = tk.Scrollbar(self.windows_frame, orient="vertical")
        self.scrollbar_x = tk.Scrollbar(self, orient="horizontal")
        self.windows_listbox = tk.Listbox(
            self.windows_frame, selectmode='single', highlightthickness=0,
            xscrollcommand=self.scrollbar_x.set,
            yscrollcommand=self.scrollbar_y.set,
            bg=self.enum_normal_color, fg=self.text_normal_color,
            selectbackground=self.entry_highlighted_color,
            selectforeground=self.text_highlighted_color,)

        # set up the scrollbars
        self.scrollbar_x.config(command=self.windows_listbox.xview)
        self.scrollbar_y.config(command=self.windows_listbox.yview)

        # set up the keybindings
        self.windows_listbox.bind('<Return>', self.select)
        self.scrollbar_x.bind('<Return>', self.select)
        self.scrollbar_y.bind('<Return>', self.select)
        self.windows_listbox.bind('<Double-Button-1>', self.select)
        self.ok_button.bind('<Return>', self.select)
        self.cancel_button.bind('<Return>', self.destroy)
        self.bind('<Escape>', self.destroy)

        # store the windows by title
        windows_by_title = {}
        for w in self.app_root.tag_windows.values():
            windows_by_title[w.title()] = w

        # populate the listbox
        for title in sorted(windows_by_title):
            self.list_index_to_window.append(windows_by_title[title])
            self.windows_listbox.insert('end', title)

        self.windows_listbox.select_set(0)

        # pack everything
        self.ok_button.pack(padx=12, pady=5, side='right')
        self.cancel_button.pack(padx=12, pady=5, side='left')
        self.ok_frame.pack(side='left', fill='x', expand=True)
        self.cancel_frame.pack(side='right', fill='x', expand=True)

        self.windows_listbox.pack(side='left', fill="both", expand=True)
        self.scrollbar_y.pack(side='left', fill="y")

        self.windows_frame.pack(fill="both", expand=True)
        self.scrollbar_x.pack(fill="x")
        self.button_frame.pack(fill="x")

        self.transient(self.app_root)
        self.ok_button.focus_set()
        self.grab_set()

    def destroy(self, e=None):
        try:
            self.app_root.tag_window_manager = None
        except AttributeError:
            pass
        tk.Toplevel.destroy(self)

    def select(self, e=None):
        indexes = [int(i) for i in self.windows_listbox.curselection()]
        w = self.list_index_to_window[indexes[0]]

        self.destroy()
        self.app_root.select_tag_window(w)
